#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# $Id: young.py,v 1.2 2003/07/15 09:36:58 etale Exp $
# =====================================================

"""library for Young tableaux"""

import sys
import copy
import operator
import itertools

# young
import partition
import combination
import mathformat
import iterator


__all__ = [ 'young', ]

# cache tableaux
# {key : value} = {partition : tableaux}
_cache = {}

class Young(list):
    '''class for Young
    '''
    def __repr__(self):
        return mathformat.pprint_tableaux(self)
    
    __str__ = __repr__

class YoungGenerator(list):
    '''generate Young Tableaux
    Based on partitions of N,
    generate all Young diagrams
    '''
    # 自然数N の分割数(partition)を元にして、
    # Young 図形を作成する。


    __slots__ = (
            'part', # partition
            'size', # size
            'tableaux',
            )
                 
    def __init__(self, part, size=0):
        self.part= part[:]  # part := partition

        if not size:
            # (3,2,1,1) -> 3 + 2 + 1 + 1 = 7
            self.size = sum(part)
        else:
            self.size = size

        self.tableaux = self.tableaux_initializer()

    def tableaux_initializer(self):
        '''Based on a partition, create the container of the tableaux
        '''
        # (3,1,1) -> [[None, None, None],[None], [None]]
        return [[None]*num for num in self.part]

    def get_empty_space(self, tableaux):
        '''Return empty space where a new number can be put in.
        '''
        prev_pos = self.size

        for i, row in enumerate(tableaux):
            for j, pos in enumerate(row):
                if pos:
                    # ■■
                    # go to the next col
                    continue    # １つ右へ
                else:
                    if j >= prev_pos:
                        # ■□
                        # ■□
                        pass
                    else:
                        # ■■
                        # ■□

                        prev_pos = j
                        yield (i,j)
                    # go to the next row
                    break   # 次(下)の行に移る。

    def __repr__(self):
        buff = []
        for dia in self:
            buff.append(mathformat.pprint_tableaux(dia))

        return "\n".join(buff)

    __str__ = __repr__

    def set_tableaux(self):
        '''set tableaux
        '''
        for dia in self.generate():
            self.append(Young(dia))


    def generate(self, tableaux = [], index = 1):
        '''generate tableaux
        '''

        if not tableaux:
            tableaux = self.tableaux_initializer()

        for candidate in self.get_empty_space(tableaux):

            col, row = candidate

            # set the number
            tableaux[col][row] = index

            if index == self.size:
                yield copy.deepcopy(tableaux)

            else:
                for dia in self.generate(tableaux[:], index +1):
                    yield dia

            tableaux[col][row] = None


class YoungTableaux(dict, iterator.Iterator):
    '''This class holds a sequence of Tableaux
    '''
    __slots__ = (
            'total',
            'square',
            'seq_of_partition',
            )

    #total = 0
    #square = 0

    def __init__(self, seq_of_partition):
        self.seq_of_partition = seq_of_partition[:]
        self.total  = 0
        self.square = 0
        
    def __iter__(self):
        # NOTE
        # __iter__ just returns dictionary's values without sorting,
        # so the order is not always the same.
        return itertools.chain(*self.values())

    def set_tableaux(self):
        '''set the tableaux
        '''
        for part in self.seq_of_partition:
            try:
                self[part] = _cache[part]
            except KeyError, e:
                yd = YoungGenerator(part)
                yd.set_tableaux()
                _cache[part] = yd
                self[part]   = yd

    def size(self):
        '''Return the total number of the tableaux
        '''
        # 標準ヤング盤が合計何個あるのか返す

        # TODO
        # もう少しましな名前にする

        if self.total:
            return self.total
        else:
            total = 0

        for k in self.iterkeys():
            total += len(self[k])

        self.total = total
        return total

    __len__ = size

    def report(self):
        '''display the statistical info about the tableaux

        # sample
        partition             number  square
        ----------------------------------------
        (4)                        1       1
        (3,1)                      3       9
        (2,2)                      2       4
        (2,1,1)                    3       9
        (1,1,1,1)                  1       1
        ----------------------------------------
        total                     10      24

        '''
        seq_of_part = self.keys()
        seq_of_part.sort()
        seq_of_part.reverse()

        print '%-20s%8s%8s'%('partition', 'number', 'square')
        print "-" * 40

        # initialize
        self.square = self.total = 0
        total = squares = 0

        for part in seq_of_part:
            num       = len(self[part])
            total    += num
            square    = num * num
            squares  += square
            print '%-20s%8d%8d'%(mathformat.pprint_partition(part).ljust(20), num, square)

        self.square = squares
        self.total  = total
        print "-" * 40
        print '%-20s%8d%8d'%('total', self.total, self.square)

    def display(self):
        '''Display all the tableaux
        '''
        print self

    def __repr__(self):
        return self._formatter()

    __str__ = __repr__

    def _formatter(self):
        '''formatter for tableaux
        '''
        buff = []

        parts = self.keys()
        parts.sort()
        parts.reverse()

        for part in parts:
            buff.append(`part`)

            for tableaux in  self[part]:

                buff.append(mathformat.pprint_tableaux(tableaux))

        return "\n".join(buff)


def ispartition(iterable):
    '''tests whether or not iterable is a valid partition
    '''
    for i in range(len(iterable)):
        try:
            assert isinstance(iterable[i], int), "each element must be integer"
            assert iterable[i] >= iterable[i+1], "not well ordered"
        except IndexError, e:
            break
        except AssertionError, e:
            print >> sys.stderr, "invalid partition:", iterable
            raise AssertionError, e

    return True

def young_of_sequence(part):
    '''分割が part のヤング盤をすべて返す
    '''
    # Return all Young tableaux of the partition

    # check if part is a valid partition
    ispartition(part)
    part = (part,)
    yg = YoungTableaux(part)
    yg.set_tableaux()
    return yg

def young_of_number(number):
    '''自然数 n の分割を元にしたヤング盤をすべて返す
    '''
    # Return all Young Tableaux of partitions of N

    pat = partition.partition(number)
    yg = YoungTableaux(pat)
    yg.set_tableaux()
    return yg

def young(*arg):
    '''Return all the standard Young tableaux.

    if arg is a number,
    return all the tableaux of partition(the size of a partition is the number)

    if arg is a partition,
    return all the tableaux of the partition.
    '''
    # XXX
    # there're three ways to call young.
    # [1]
    # young(3)       # arg = (number,)
    # [2]
    # young((2,1,1)) # arg = (partition,)
    # [3]
    # young(2,1,1)   # arbitrary number of arguments
    #
    # I think the third one is much more intuitive than the second one.

    if len(arg) == 1:
        arg = arg[0]

    if isinstance(arg, int):
        return young_of_number(arg)
    elif isinstance(arg, list) or isinstance(arg, tuple):
        return young_of_sequence(arg)
    else:
        raise TypeError, "argument must be integer or partition"

    return


# TEST
def _test1():
    # calculate tableaux based on the partition

    p = (2,1,1)
    y = young(p)
    print y
    print "len       : %d"%len(y)
    print "size      : %d"%y.size()

    return

def _test2():
    # calculate tableaux based on the number

    number = 4
    y = young(number)
    # test of __repr__ in Young class
    for tbl in y:
        print tbl

    y.report()
    print "factorial of %d : %d"%(number, combination.factorial(number))

def _test3():
    # use arbitrary argument lists
    # test for iterator module

    y = young(5,1,1)
    print y

    print "take(4)"
    y.take(4)
    print "drop(10)"
    y.drop(10)
    print "last(3)"
    y.last(3)


if __name__ == '__main__':
    _test1()
    _test2()
    _test3()

